# ========================================================================================
# Purpose: Python script for calling CMRJack from SPSS
# File: CALLJACK.py
# Written by: Jon Pedersen
# Date: March 1 2015
# Note: Should be located in the IBM SPSS extension folder together with CALLJACK.xml
#       Does not work independently
# Revision: 12.07.2019
#           Added support for stillbirths
#           24.11.2019
#           Fixed error in output encoding
#           22.05.2020
#           Removed requirement and output of current age for FBH (not used)
#           19/04/2021
#			Update for python 3.8+, command print, args are changed 
# ========================================================================================
#
# Parameter list:
#
# Imports
import spss, spssaux
import sys
import string
import tempfile
import random
from extension import Template, Syntax, processcmd
from subprocess import call
from winreg import *
import os


ScriptVersion="Version 1.2 May, 2020"
HelpText = "Help for CALLJACK " + ScriptVersion + "\n" + r"""
  CALLJACK
    [/OUTPUT FILE  = {file}]
    /SAMPLE WEIGHT = {Var} STRATUM = {Var} CLUSTER = {Var} DATEINT = {Var}
    [/FBH BIRTHDATE = {Var} ALIVE = {Var} AGEDEATH = {Var}]
      [PERIODS={list of 1,2,3,4,5**,10}]
      [OPTIMAL = {YES|NO**}]
      [STILLBIRTHS = {NONE**|SBONLY|ALL}]
      [TIME = {CALENDAR**|BEFORESURVEY}]
    [/SBHAGE AGEWOMAN = {Var} DEADBOYS = {Var}DEADGIRLS = {Var} LIVEBOYS = {Var} LIVEGIRLS = {Var}]
    [/SBHTSFB TSFB = {Var} DEADBOYS = {Var} DEADGIRLS = {Var} LIVEBOYS = {Var} LIVEGIRLS = {Var}]
    [/LIFETABLE MODEL=[UNLATINAMERICAN | UNCHILEAN | UNSOUTHASIAN | UNFAREAST |UNGENERAl | NORTH | EAST |SOUTH | WEST **]
            [PRINT = {SELECTEDMODEL** | ALLMODELS}]]
    [/FACTOR VARIABLE= {Var} TYPE={SIMPLE**|RATIO|INEQUITY}]
    [/TRANSFORM SCALE = {REAL** | LOG| LOGIT}]
    [/VARIANCE ESTIMATOR={STRATIFIED**|NOSTRATA}]
    [/SURVEY [NAME={text}] [YEAR={text}] [TYPE={text}] [COUNTRY={text}]]
    [/PRINT FORMAT=[PRESENTATION] [CMEINFO] [RAW**] [OPTIMAL] [CEBCSTABLE][DOCUMENTATION]]
    [/HELP].

    ** Default if the subcommand is omitted.

    OUTPUT controls the filename of the Excel file with results

    SAMPLE describes the sample structure with three variables WEIGHT STRATUM and CLUSTER
    and also the Date of interview (DATEINT)

    Use only one of the commands
    FBH Full birth history.
    Requires a list of variables that provide birthdate of the child, gender, whether the child is currently
    living, the age of death, and the current age.
    PERIODS allows you to supply the periods required for estimation
    OPTIMAL turns calculation of optimal periods on (OPTIMAL=YES) or off (OPTIMAL=NO).
    STILLBIRTHS  controls generation of fetal loss estimates NONE means no estimates (even if data are present)
    SBONLY means estimation of stillbirth rates only, while all estimates  stillbirths, miscarriages and other terminations.

    SBHAGE Summary birth history with classication after age of mother.
    The subcommand requires a list of variables:
    AGEWOMAN, DEADBOYS, DEADGIRLS, LIVEBOYS, LIVEGIRLS

    SBTSFB Summary birth history with classification after time since first birth
    Summary birth history with classication after age of mother.
    The subcommand requires a list of variables: ,
    TSFB, DEADBOYS, DEADGIRLS, LIVEBOYS, LIVEGIRLS

    LIFTEABLE specifies the lifetable model.
    LIFETABLE are only used in the context of summary birth histories.
    Summary birth histories using the Time Since First Birth approach (SBTSFB) can only use the Coale-Demeny
    North, East, South and West.
    The categories variables allows for defining the standard MICS breakdown variables.

    FACTOR allow you to supply a variable that is used to provide subgroups in the tabulation.
    The keyword SIMPLE to the FACTOR command means that simple estimates should be produce; RATIO means that ratios based
    on the variable should be produced and INEQUITY produces bivariate inequity measures.

    TRANSFORM and its subcommand SCALE determine if the estimates and their standard error should be on the REAL scale,
    the LOG scale or the LOGIT scale.

    SURVEY allows you to give a title to the output.
     example SURVEY NAME="DR Congo" YEAR="2013-14" TYPE="DHS" COUNTRY="Democratic Republic of Congo"

    PRINT controls the output. Choices of FORMAT is PRESENTATION (easily readable tables)
     CMEINFO (Output used to input into the CME Info database - little use for other users)
     RAW (File with the estimates, each as a column)
     CEBCSTABLE (Table with estimates of children ever born, children surviving)
    """
def VarIsThere(Var):
    if Var is None:
        b = False
    elif Var == "":
        b = False
    else:
        b = True
    return b

def randomsavname():
    return 'A$' + '{0}'.format(random.randrange(100000, 999999))+'.sav'

def Run(args):
    args = args[list(args.keys())[0]]
    oobj = Syntax([
        Template("FILE", subc="OUTPUT", ktype="literal", var="Outputfile"),
        Template("WEIGHT", subc="SAMPLE", ktype="varname", var="WeightVar"),
        Template("STRATUM", subc="SAMPLE", ktype="varname", var="StratumVar"),
        Template("CLUSTER", subc="SAMPLE", ktype="varname", var="ClusterVar"),
        Template("DATEINT", subc="SAMPLE", ktype="varname", var="DateIntVar"),
        Template("BIRTHDATE", subc="FBH", ktype="varname", var="BirthDateVar"),
        Template("SEX", subc="FBH", ktype="varname", var="SexVar"),
        Template("ALIVE", subc="FBH", ktype="varname", var="AliveVar"),
        Template("AGEDEATH", subc="FBH", ktype="varname", var="AgeDeathVar"),
        Template("CURRAGE", subc="FBH", ktype="varname", var="CurrageVar"),
        Template("OPTIMAL",subc="FBH",ktype="literal", var="OptVar"),
        Template("TIME",subc="FBH",ktype="literal", var="TimeVar"),
        Template("STILLBIRTHS",subc="FBH",ktype="literal", var="SBVar"),
        Template("PERIODS", subc="FBH",ktype="literal",islist=True,var="PeriodsVar"),
        Template("AGEWOMAN", subc="SBHAGE", ktype="varname", var="AoWVar"),
        Template("DEADBOYS", subc="SBHAGE", ktype="varname", var="DeadBoysVar"),
        Template("DEADGIRLS", subc="SBHAGE", ktype="varname", var="DeadGirlsVar"),
        Template("LIVEBOYS", subc="SBHAGE", ktype="varname", var="LiveBoysVar"),
        Template("LIVEGIRLS", subc="SBHAGE", ktype="varname", var="LiveGirlsVar"),
        Template("TSFB", subc="SBHTSFB", ktype="varname", var="TSFBVar"),
        Template("DEADBOYS", subc="SBHTSFB", ktype="varname", var="DeadBoysVar"),
        Template("DEADGIRLS", subc="SBHTSFB", ktype="varname", var="DeadGirlsVar"),
        Template("LIVEBOYS", subc="SBHTSFB", ktype="varname", var="LiveBoysVar"),
        Template("LIVEGIRLS", subc="SBHTSFB", ktype="varname", var="LiveGirlsVar"),
        Template("MODEL", subc="LIFETABLE", ktype="literal", var="ModelVar"),
        Template("VARIABLE", subc="FACTOR", ktype="varname", var="MainVar"),
        Template("TYPE",subc="FACTOR",ktype='literal',var="MainVarType"),
        Template("SCALE",subc="TRANSFORM",ktype='literal',var="LogVar"),
        Template("ESTIMATOR",subc="VARIANCE",ktype="literal", var="VarVar"),
        Template("NAME", subc="SURVEY",ktype="literal",var="SurveyVar"),
        Template("YEAR",subc="SURVEY",ktype="literal",var="SurveyYearVar"),
        Template("COUNTRY",subc="SURVEY",ktype="literal",var="SurveyCountryVar"),
        Template("TYPE",subc="SURVEY",ktype="literal",var="SurveyTypeVar"),
        Template("FORMAT",subc="PRINT",ktype='literal',islist=True,var="PrintVar"),
        Template("HELP", subc="", ktype="bool")])
    # ---
    # A HELP subcommand overrides all else
    if "HELP" in args:
        print ( HelpText )
    else:
        processcmd(oobj, args, CallJACK)

def CallJACK(Outputfile=None,WeightVar=None,StratumVar=None,ClusterVar=None,DateIntVar=None,BirthDateVar=None,SexVar=None,
             AliveVar=None,AgeDeathVar=None,CurrageVar=None,OptVar="NO", TimeVar="CALENDAR", SBVar="NONE",PeriodsVar=[5], AoWVar=None,DeadBoysVar=None, 
             DeadGirlsVar=None,
             LiveBoysVar=None,LiveGirlsVar=None,TSFBVar=None,ModelVar="WEST",MainVar=None,MainVarType="SIMPLE",LogVar="REAL",
             VarVar="STRATIFIED", SurveyVar=None,SurveyYearVar=None,
             SurveyCountryVar=None,SurveyTypeVar=None,PrintVar=[u"RAW"]):
    #
    # Build required export file from spss
    # To do that, find out what variables are used, and what method is required
    # at the same time build command line to be passed to CMRJack
    # file, designvars, type of analysis, variables, options
    #
    # print PrintVar
    if VarIsThere(Outputfile):
        #Outputfile.replace("/","\\")
        cline = '/o:"' +Outputfile+'"'
    else:
        cline = '"/o:cmrjackout.xlsx"'
        print ( "Outputfile omitted, cmrjackout.xlsx used" )

    cline += " /1:" + WeightVar + " /2:"+StratumVar + " /3:" + ClusterVar + " /4:" + DateIntVar
    KeepString = WeightVar+" "+StratumVar+" "+ClusterVar + " " + DateIntVar
    if VarIsThere(BirthDateVar):
        cline += " /a:FBH" + " /5:"+BirthDateVar+" /6:"+AliveVar+" /7:"+AgeDeathVar
        #+" /8:"+CurrageVar
        #FBHVars=" "+BirthDateVar+" "+SexVar+" "+AliveVar+" "+AgeDeathVar+" "+CurrageVar
        KeepString += " "+BirthDateVar+" "+AliveVar+" "+AgeDeathVar
        #+" "+CurrageVar
    elif VarIsThere(AoWVar):
        cline += " /a:AOW" + " /9:"+AoWVar+" /11:"+DeadBoysVar+" /12:"+DeadGirlsVar+" /13:"+LiveBoysVar+" /14:"+LiveGirlsVar
        AOWVars = " "+AoWVar+" "+DeadBoysVar+" "+DeadGirlsVar+" "+LiveBoysVar+" "+LiveGirlsVar
        KeepString += AOWVars
    elif VarIsThere(TSFBVar):
        cline += " /a:TSFB" + " /10:"+TSFBVar+" /11:"+DeadBoysVar+" /12:"+DeadGirlsVar+" /13:"+LiveBoysVar+" /14:"+LiveGirlsVar
        TSFBVars = " "+TSFBVar+" "+DeadBoysVar+" "+DeadGirlsVar+" "+LiveBoysVar+" "+LiveGirlsVar
        KeepString += TSFBVars
        if VarIsThere(ModelVar):
            if ModelVar[:2] == "UN":
                print ( """UN family of model life tables cannot be used for Time Since First Birth
                summary birth histories. Use NORTH, EAST, SOUTH or WEST or let CMRJack choose
                by not specifying the LIFETABLE option""" )
                quit()
    else:
        print ( "Either FBH, SBHAGE or SBHTSFB must be specified" )
        quit()
    if VarIsThere(MainVar):
        KeepString += ' ' + MainVar
        cline+=" /FA:" + MainVar
        cline+=" /FT:" + MainVarType
    bProduceOptimal=False
    if VarIsThere(OptVar):
        cline+= " /OPT:" + OptVar
        if OptVar=="YES":
            bProduceOptimal=True
    if VarIsThere(SBVar):
        cline+= " /SB:" + SBVar    

    if VarIsThere(PeriodsVar):
        textRep=""
        for a in PeriodsVar:
            textRep+=str(a)+" "
        cline+= " /P:"+'"'+textRep.strip()+'"'

    if VarIsThere(LogVar):
        cline+= " /L:" + '"'+LogVar+'"'

    if VarIsThere(SurveyVar):
        cline+= " /SV:" + '"'+SurveyVar+'"'

    if VarIsThere(SurveyYearVar):
        cline+=" /SVY:"+ '"'+SurveyYearVar+'"'

    if VarIsThere(SurveyCountryVar):
        cline+=" /SVC:"+'"'+SurveyCountryVar+'"'

    if VarIsThere(SurveyTypeVar):
            cline+=" /SVT:"+'"'+SurveyTypeVar+'"'

    if VarIsThere(PrintVar):
        if bProduceOptimal:
            n=16
        else:
            n=0
        for a in PrintVar:
            if a=="PRESENTATION":
                n+=2
            elif a=="CMEINFO":
                n+=4
            elif a== "RAW":
                n+=8
            elif a== "CEBCSTABLE":
                n+=16
            elif a== "OPTIMAL" and not bProduceOptimal:
                n+=32    
            elif a== "DOCUMENTATION":
                n+=64
        cline+= " /PT:"+str(n)

    # Finished with building variable list, sort, save temporary sav-file in temp-directory
    spss.Submit('sort cases by {0} {1}.'.format(StratumVar, ClusterVar))
    savfilename =  tempfile.gettempdir() + '\\' + randomsavname()
    spss.Submit("save outfile = '{0}' /KEEP {1}.".format(savfilename, KeepString))
    cline += ' /s:"{0}"'.format(savfilename)
    # Build command line to CMRJack
    # This is basically list of variables and options
    #
    # call CMRJack, but first find out where it is (it stores its location in the registry)
    aReg = ConnectRegistry(None,HKEY_CURRENT_USER)
    aKey = OpenKey(aReg, r"SOFTWARE\CMRJack\CMRJack")
    val=QueryValueEx(aKey, "ExeDir")
    call(val[0]+' '+cline)
    #print cline
    val=QueryValueEx(aKey, "LatestFileSaved")
    spss.StartProcedure("CMRJack")
    textBlock = spss.TextBlock("Processing finished. ","Saved output in "+val[0])
    spss.EndProcedure()

